# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
The following functions relate to management of NVSHMEM symmetric memory in Python
"""
import logging

from nvshmem.core.nvshmem_types import *
import nvshmem.bindings as bindings
from nvshmem.core._internal_tracking import _mr_references, _cached_device, _is_initialized, InternalInitStatus
from nvshmem.core.utils import _get_device

from cuda.core import Device, system
from cuda.core import Buffer

__all__ = [
    'buffer', 'free', 'get_peer_buffer', 'get_multicast_buffer', 'register_external_buffer',
    'unregister_external_buffer'
]

logger = logging.getLogger("nvshmem")


def _free_all_buffers() -> None:
    """
    Frees all allocated NVSHMEM buffers currently being tracked.

    This is an internal helper used to clean up all memory allocations managed by
    ``_mr_references``. Each allocation is freed only once its reference count has reached zero.

    Logs each deallocation for debugging or auditing purposes.
    """
    if not _is_initialized["status"] == InternalInitStatus.INITIALIZED:
        logger.warning("NVSHMEM Library is not initialized. Cannot free buffers")
        return
    found_leak = False
    for key in sorted(_mr_references.keys()):
        mr = _mr_references[key]
        for ptr in sorted(mr._mem_references.keys()):
            if mr._mem_references[ptr]["type"] != BufferTypes.NORMAL or mr._mem_references[ptr]["freed"]:
                continue
            found_leak = True
            logger.info(
                f"Found object open at pointer {ptr} and ref count {mr._mem_references[ptr]['ref_count']}. Freeing it.")
            # We already printed the warning message so we can safely suppress the message
            mr._mem_references[ptr]["freed"] = True
            mr._mem_references[ptr]["released"] = True
            mr._mem_references[ptr]["except_on_del"] = False
            mr._mem_references[ptr]["ref_count"] = 0
            mr.deallocate(ptr, 0)
    if found_leak:
        logger.error(
            "Some NVSHMEM Symmetric memory was not freed explicitly (you may have forgotten to clean up before finalizing, or an unrelated exception crashed the program before it was freed)"
        )


def buffer(size, release=False, except_on_del=True) -> Buffer:
    """
    Allocates an NVSHMEM-backed CUDA buffer.

    Args:
        size (int): The size in bytes of the buffer to allocate.
        release (bool, optional): Do not track this buffer internally to NVSHMEM
                If True, it is the user's responsibility to hold references to the buffer until free() is called
                otherwise, deadlocks may occur.
        except_on_del (bool, optional): If True, raise an exception if the buffer is not tracked or has already been freed.
                If False, print an error message.

    Returns:
        ``cuda.core.Buffer``: A DLPack-compatible CUDA buffer with NVSHMEM backing.

    Raises:
        ``NvshmemError``: If the buffer could not be allocated properly.

    Note that this is a collective. All participating PEs must call ``buffer()`` in concert.

    This operation runs on the cached Device. If the cached Device is not the current device, it will set the cached device to current and set it back at the end of the operation.
    """
    if _is_initialized["status"] != InternalInitStatus.INITIALIZED:
        raise NvshmemInvalid("NVSHMEM Library is not initialized")

    user_nvshmem_dev, other_dev = _get_device()

    dev_id = user_nvshmem_dev.device_id
    resource = _mr_references.get(dev_id)
    if resource is None:
        logger.debug(f"Creating NvshmemResource for device {dev_id}")
        resource = NvshmemResource(user_nvshmem_dev)
        _mr_references[dev_id] = resource

    buf = resource.allocate(size, release=release, except_on_del=except_on_del)

    if other_dev is not None:
        other_dev.set_current()
    return buf


def free(buffer: Buffer) -> None:
    """
    Frees an NVSHMEM buffer that was previously allocated.

    Args:
        buffer (``cuda.core.Buffer``): The buffer to free.

    Raises:
        - ``NvshmemInvalid``: If the buffer is not a valid NVSHMEM-managed buffer.
        - ``NvshmemError``: If the buffer is not tracked or has already been freed.

    Note that this is a collective. All participating PEs must call ``free()`` in concert.
    """
    if _is_initialized["status"] != InternalInitStatus.INITIALIZED:
        logger.warning("NVSHMEM Library is not initialized. Cannot free buffer")
        return
    # _get_device() excepts if no device is current
    user_nvshmem_dev, other_dev = _get_device()

    if not (isinstance(buffer, Buffer)) or not hasattr(buffer, "handle"):
        raise NvshmemInvalid("Tried to free a buffer not from NVSHmem")

    buffer.memory_resource.set_freed(buffer)
    try:
        buffer.memory_resource.set_freed(buffer)
    except NvshmemError:
        logger.error(f"Freed a buffer {buffer} that was previously freed or not tracked")

    buffer.close()
    if other_dev is not None:
        other_dev.set_current()


def get_peer_buffer(buffer: Buffer, pe: int):
    """
    Returns a peer buffer associated with an NVSHMEM-allocated object.

    This is the Python object equivalent of nvshmem_ptr, which:
        - Given a pointer to an object on the NVSHMEM Symmetric Heap
        - Returns a pointer to a local object to which loads and stores can be performed

    The Python equivalent returns a cuda.core.Buffer which starts at the address of the Buffer passed in, with the same size as the Buffer passed in.

    For more information on nvshmem_ptr, see https://docs.nvidia.com/nvshmem/archives/nvshmem-101/api/docs/gen/api/setup.html#nvshmem-ptr

    The get_peer_buffer function offers an efficient means to accomplish communication, for example when a sequence of reads and writes to a data object on a remote PE does not match the access pattern provided in other APIs.

    Args:
        - buffer (``cuda.core.Buffer``): A buffer allocated with NVSHMEM.
        - pe (``int``): The peer's PE

    Returns:
        - ``cuda.core.Buffer``: The buffer object representing the remote peer's buffer.
            User need not call ``nvshmem.core.free()`` on this Buffer. It will be a no-op

    Raises:
        - ``NvshmemInvalid``: If the input buffer is not a valid NVSHMEM buffer.
        - ``NvshmemError``: If the buffer is not tracked internally or no peer information is found.
    """
    if _is_initialized["status"] != InternalInitStatus.INITIALIZED:
        raise NvshmemInvalid("NVSHMEM Library is not initialized")

    # _get_device() excepts if no device is current
    user_nvshmem_dev, other_dev = _get_device()

    if not isinstance(buffer, Buffer) or not hasattr(buffer, "handle"):
        raise NvshmemInvalid("Tried to use a buffer not from NVSHmem")

    mr = buffer.memory_resource
    peer_buffer = mr.get_peer_buffer(buffer, pe)
    if other_dev is not None:
        other_dev.set_current()
    return peer_buffer


def register_external_buffer(buffer: Buffer) -> Buffer:
    """
    Register an external buffer with NVSHMEM.
    
    This function is a collective. All participating PEs must call ``register_external_buffer()`` in concert.
    This function will return a new buffer object that wraps a symmetric pointer to the external buffer.
    Users must pass this buffer to NVSHMEM operations, rather than the original buffer.

    Args:
        - buffer (``cuda.core.Buffer``): A buffer to register with NVSHMEM.

    Returns: 
        - ``cuda.core.Buffer``: A buffer object wrapping the registered external buffer.

    The user must call ``unregister_external_buffer()` on the buffer returned by this 
    function to unregister the buffer from NVSHMEM when they are done with it.
    """
    if _is_initialized["status"] != InternalInitStatus.INITIALIZED:
        raise NvshmemInvalid("NVSHMEM Library is not initialized")

    # _get_device() excepts if no device is current
    user_nvshmem_dev, other_dev = _get_device()

    # Unlike other functions, we do not expect the buffer to be tracked by NVSHMEM
    mr = _mr_references.get(user_nvshmem_dev.device_id)
    if mr is None:
        raise NvshmemInvalid("Tried to register an external buffer on a device that is not initialized with NVSHMEM")
    registered_buffer = mr.register_external_buffer(buffer)
    if other_dev is not None:
        other_dev.set_current()
    if registered_buffer is not None:
        return registered_buffer


def unregister_external_buffer(buffer: Buffer) -> None:
    """
    Unregister an external buffer with NVSHMEM.

    This function is a collective. All participating PEs must call ``unregister_external_buffer()`` in concert.
    
    Args:
        - buffer (``cuda.core.Buffer``): The buffer returned by ``register_external_buffer()`` to unregister.
    
    Raises:
        - ``NvshmemInvalid``: If NVSHMEM is not initialized or buffer is invalid.
    """
    if _is_initialized["status"] != InternalInitStatus.INITIALIZED:
        raise NvshmemInvalid("NVSHMEM Library is not initialized")

    # _get_device() excepts if no device is current
    user_nvshmem_dev, other_dev = _get_device()

    if not isinstance(buffer, Buffer) or not hasattr(buffer, "handle"):
        raise NvshmemInvalid("Tried to use a buffer not from NVSHmem")

    mr = _mr_references.get(user_nvshmem_dev.device_id)
    if mr is None:
        raise NvshmemInvalid("Tried to unregister an external buffer on a device that is not initialized with NVSHMEM")
    mr.unregister_external_buffer(buffer)
    if other_dev is not None:
        other_dev.set_current()


def get_multicast_buffer(team: Teams, buffer: Buffer) -> Buffer:
    """
    Returns a peer buffer associated with an NVSHMEM-allocated object which uses Multicast Memory.

    This is the Python object equivalent of ``nvshmemx_mc_ptr``, which:
        - Given a pointer to an object on the NVSHMEM Symmetric Heap
        - Returns a pointer to a local object to which multicast loads and stores can be performed

    The Python equivalent returns a cuda.core.Buffer which starts at the address of the Buffer passed in, with the same size as the Buffer passed in.

    For more information on ``nvshmemx_mc_ptr``, see https://docs.nvidia.com/nvshmem/api/using.html#communication-model, or
      https://docs.nvidia.com/nvshmem/api/gen/api/setup.html?highlight=mc_ptr#c.nvshmemx_mc_ptr

    The ``get_multicast_buffer`` function offers a quick means to leverage NVSwitch based multicast and reduction offload features also referred to as NVLink SHARP.

    IMPORTANT: The buffer allocated by ``get_multicast_buffer`` may NOT be accessed from the host (CPU). It can only be accessed by GPU kernels.

    NOTE: At present, copy_from() and copy_to() are not supported for Multicast buffers.

    Args:
        - buffer (``cuda.core.Buffer``): A multicast buffer allocated with NVSHMEM.
        - Team (``Teams`` enum): The team which this multicast memory applies to

    Returns:
        - ``cuda.core.Buffer``: The buffer object representing the remote peer's buffer.
            User need not call ``nvshmem.core.free()`` on this Buffer. It will be a no-op

    Raises:
        - ``NvshmemInvalid``: If the input buffer is not a valid NVSHMEM buffer.
                              or if the platform does not support Multicast memory
        - ``NvshmemError``: If the buffer is not tracked internally or no team information is found.
    """
    if _is_initialized["status"] != InternalInitStatus.INITIALIZED:
        raise NvshmemInvalid("NVSHMEM Library is not initialized")

    # _get_device() excepts if no device is current
    user_nvshmem_dev, other_dev = _get_device()

    if not isinstance(buffer, Buffer) or not hasattr(buffer, "handle"):
        raise NvshmemInvalid("Tried to use a buffer not from NVSHmem")

    mr = buffer.memory_resource
    peer_buffer = mr.get_mc_buffer(team, buffer)
    if other_dev is not None:
        other_dev.set_current()
    return peer_buffer



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
The following are utility functions for NVSHMEM4Py
"""

import logging
import sys
import os
import socket
import threading
from enum import IntEnum

import numpy as np

import nvshmem.bindings as bindings
from nvshmem.core._internal_tracking import _mr_references, _cached_device, _debug_mode

from cuda.core import Device, system
from cuda.core import Buffer

try:
    import torch
    torch_to_numpy = {
        torch.float16: np.float16,
        torch.bfloat16: np.float16,  # Note: bfloat16 isn't supported by default in NumPy
        torch.float32: np.float32,
        torch.float64: np.float64,
        torch.int8: np.int8,
        torch.uint8: np.uint8,
        torch.uint16: np.uint16,
        torch.uint32: np.uint32,
        torch.uint64: np.uint64,
        torch.int16: np.int16,
        torch.int32: np.int32,
        torch.int64: np.int64,
        torch.bool: np.bool_,
    }
    _torch_enabled = True
except:
    torch_to_numpy = None
    _torch_enabled = False


def _get_device() -> Device:
    """
    Perform a pre-flight check on the cuda Device in use, making sure it's ready for use in nvshmem
    This function returns the Device that is current before it was called, and sets the current device 
    to the cached device that this NVSHMEM PE was registered against at init time.

    The caller MUST set the same device back to current at the end of the API calling this.

    This function raises an Exception if no device is set to current.

    The expected, required flow for an NVSHMEM4Py.Core API is:
        Call ``_get_device()``, save old device
                internally, this sets current device to the correct device this PE was initialized with
        Perform API call
        Set old device back to current. If the old device is the same as the current one, this is idempotent
        If not, it matches the NVSHMEM4Py and Libnvshmem contract - return the process to the caller as they gave it to us
    """
    if _debug_mode:
        # Device() excepts if no device is current
        old_device = Device()
    else:
        old_device = None

    if not _debug_mode and _cached_device["device"] is not None:
        return _cached_device["device"], None

    # If the device is None here, we want to do the second half of the two-stage init.
    if _cached_device["device"] is None:
        _cached_device["device"] = Device()

    if _cached_device["device"].device_id != Device().device_id:
        _cached_device["device"].set_current()

    return _cached_device["device"], old_device


def _configure_logging(level="WARNING", logfile=None, mype=None):
    """
    Configure the Python logger for NVSHMEM or related modules.

    NOTE: mype can be retrieved with ``nvshmem.core.my_pe()``

    Args:
        level (str): Logging level as a string. Options are 'DEBUG', 'INFO', 
                     'WARNING', 'ERROR', or 'CRITICAL'. Defaults to 'WARNING'.
        logfile (str, optional): If provided, logs will also be written to this file.

    Example:
        >>> configure_logging("DEBUG")
        >>> logging.debug("This is a debug message.")
    """
    if mype is None:
        try:
            mype = bindings.my_pe()
        except:
            print("Unable to retrieve PE. Ensure NVSHMEM is initialized")
            mype = -1
    numeric_level = getattr(logging, level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f"Invalid log level: {level}")

    # C NVSHMEM Log Format: host:pid:tid [PE] NVSHMEM <log_level>
    host = socket.getfqdn().split(".")[0]
    pid = os.getpid()
    tid = threading.get_native_id()
    # Note, my preference is to include time in the log fmt, but libnvshmem doesn't do this.
    formatter = logging.Formatter(f"{host}:{pid}:{tid} [{mype}] " + "NVSHMEM %(levelname)s : %(message)s")

    # Create console handler
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(formatter)

    # Set up the root logger
    logger = logging.getLogger("nvshmem")
    logger.setLevel(numeric_level)
    logger.handlers = []  # Remove any default handlers
    logger.addHandler(console_handler)

    if logfile:
        file_handler = logging.FileHandler(logfile)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)


def get_size(shape, dtype):
    """
    Calculate the total size in bytes of an array with the given shape and dtype.

    Parameters:
        shape (tuple or list of int): Shape of the array (e.g., (64, 64)).
        
        dtype (str, np.dtype, or torch.dtype): Data type (e.g., 'float32', torch.float64).

    Returns:
        
        int: Total size in bytes.
    """
    # Normalize dtype to np.dtype

    if _torch_enabled and isinstance(dtype, torch.dtype):
        if not torch_to_numpy or dtype not in torch_to_numpy:
            raise ValueError(f"Unsupported torch dtype: {dtype}")
        dtype = np.dtype(torch_to_numpy[dtype])
    else:
        # Special case BF16 because numpy doesn't support it
        if dtype == "bfloat16":
            dtype = np.float16
        dtype = np.dtype(dtype)

    num_elements = np.prod(shape)
    return int(num_elements * dtype.itemsize)


def dtype_nbytes(dtype: str) -> int:
    """Return the size in bytes of a single element of the given NVSHMEM dtype.

    Accepts NVSHMEM dtype names such as ``'float'``, ``'int64'``, ``'half'``, etc.
    Raises a ``ValueError`` if the dtype is not recognized.

    Args:
        dtype (str): The NVSHMEM data type name.

    Returns:
        int: The number of bytes required to store a single element of the given dtype.

    Raises:
        ``ValueError``: If the specified dtype is not supported.
    """
    dtype_sizes = {
        "half": 2,
        "bfloat16": 2,
        "float": 4,
        "double": 8,
        "uint8": 1,
        "uint16": 2,
        "uint32": 4,
        "uint64": 8,
        "int8": 1,
        "int16": 2,
        "int32": 4,
        "int32_t": 4,
        "uint32_t": 4,
        "int64_t": 8,
        "uint64_t": 8,
        "long": 8,
        "longlong": 8,
        "ulonglong": 8,
        "ptrdiff": 8,
        "fp16": 2,
        "bfloat16": 2,
        "int": 4,
        "int64": 8,
        "char": 1,
        "schar": 1,
        "size": 8,  # assuming size_t is 64-bit
    }

    try:
        return dtype_sizes[dtype]
    except KeyError:
        raise ValueError(f"Unsupported NVSHMEM dtype: {dtype}")



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

__all__ = ["numba", "cute"]

# Lazy import for numba and cute - only import when accessed
_numba_module = None
_cute_module = None


def __getattr__(name):
    """Lazy import of numba and cute modules."""
    if name == "numba":
        global _numba_module
        if _numba_module is None:
            from . import numba
            _numba_module = numba
        return _numba_module
    if name == "cute":
        global _cute_module
        if _cute_module is None:
            from . import cute
            _cute_module = cute
        return _cute_module
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# Copyright (c) 2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from nvshmem.bindings.device.cute import *
from nvshmem.bindings.device.cute import (
    barrier as _nvshmem_barrier,
    barrier_block as _nvshmem_barrier_block,
    barrier_warp as _nvshmem_barrier_warp,
)

import nvshmem.core
import cutlass
from cutlass import cute
from cutlass.base_dsl.ast_helpers import const_expr
from cutlass.base_dsl.typing import cast as cute_cast

__all__ = [
    "sync_block",
    "sync_warp",
    "sync",
    "sync_all",
    "sync_all_block",
    "sync_all_warp",
    "barrier",
    "barrier_block",
    "barrier_warp",
    "barrier_all",
    "barrier_all_block",
    "barrier_all_warp",
    "reduce",
    "reduce_block",
    "reduce_warp",
    "reducescatter",
    "reducescatter_block",
    "reducescatter_warp",
    "fcollect",
    "fcollect_block",
    "fcollect_warp",
    "broadcast",
    "broadcast_block",
    "broadcast_warp",
    "alltoall",
    "alltoall_block",
    "alltoall_warp",
]


@cute.jit
def _resolve_ptr(arg):
    return arg.iterator


@cute.jit
def _size_of(obj):
    return cute.size(obj)


@cute.jit
def _resolve_dtype(dst):
    return dst.element_type


@cute.jit
def _resolve_team(team):
    return cute_cast(team, cutlass.Int32)


@cute.jit
def _resolve_root(root):
    return cute_cast(root, cutlass.Int32)


@cute.jit
def _resolve_nelems_dst(dst):
    return cute_cast(_size_of(dst), cutlass.Uint64)


@cute.jit
def _resolve_nelems_src(src):
    return cute_cast(_size_of(src), cutlass.Uint64)


@cute.jit
def _resolve_nelems_min(dst, src):
    dst_nelems = _size_of(dst)
    src_nelems = _size_of(src)
    return cute_cast(dst_nelems if dst_nelems < src_nelems else src_nelems, cutlass.Uint64)


@cute.jit
def _resolve_nelems_alltoall(src, team):
    src_nelems = cute_cast(_size_of(src), cutlass.Uint64)
    team_size = cute_cast(team_n_pes(team), cutlass.Uint64)
    return cute_cast(src_nelems // team_size, cutlass.Uint64)


# sync variations


@cute.jit
def sync_block(team):
    """
    Executes a CTA-level synchronization across all PEs in ``team``. All threads in the CTA must call this function.

    This is a lightweight synchronization point that guarantees all PEs in the team have
    reached it before any PE proceeds.  It does not provide memory-ordering or
    memory-visibility guarantees; use ``barrier`` when a memory fence is also required.
    All PEs in the team must call this function before any PE can proceed past it.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the set of PEs to synchronize.
          Use ``nvshmem.core.Teams.TEAM_WORLD`` to synchronize all PEs.

    Note:
        All PEs in ``team`` must call ``sync_block`` with the same ``team`` argument.
        Use ``sync_all_block`` to synchronize across all PEs without specifying a team.
    """
    team = _resolve_team(team)
    return team_sync_block(team)


@cute.jit
def sync_warp(team):
    """
    Executes a warp-level synchronization across all PEs in ``team``. All threads in the warp must call this function.

    This is a lightweight synchronization point that guarantees all PEs in the team have
    reached it before any PE proceeds.  It does not provide memory-ordering or
    memory-visibility guarantees; use ``barrier`` when a memory fence is also required.
    All PEs in the team must call this function before any PE can proceed past it.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the set of PEs to synchronize.
          Use ``nvshmem.core.Teams.TEAM_WORLD`` to synchronize all PEs.

    Note:
        All PEs in ``team`` must call ``sync_warp`` with the same ``team`` argument.
        Use ``sync_all_warp`` to synchronize across all PEs without specifying a team.
    """
    team = _resolve_team(team)
    return team_sync_warp(team)


@cute.jit
def sync(team):
    """
    Executes a thread-level synchronization across all PEs in ``team``.

    This is a lightweight synchronization point that guarantees all PEs in the team have
    reached it before any PE proceeds.  It does not provide memory-ordering or
    memory-visibility guarantees; use ``barrier`` when a memory fence is also required.
    All PEs in the team must call this function before any PE can proceed past it.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the set of PEs to synchronize.
          Use ``nvshmem.core.Teams.TEAM_WORLD`` to synchronize all PEs.

    Note:
        All PEs in ``team`` must call ``sync`` with the same ``team`` argument.
        Use ``sync_all`` to synchronize across all PEs without specifying a team.
    """
    team = _resolve_team(team)
    return team_sync(team)


# sync_all variations


@cute.jit
def sync_all_block():
    """
    Executes a CTA-level synchronization across all PEs in the NVSHMEM runtime
    (equivalent to ``sync_block(TEAM_WORLD)``). All threads in the CTA must call this function.

    This is a convenience wrapper around ``sync_block`` that automatically uses
    ``TEAM_WORLD`` as the team, covering all PEs participating in the NVSHMEM job.

    Note:
        All PEs must call ``sync_all_block`` before any PE can proceed past it.
    """
    team = _resolve_team(nvshmem.core.Teams.TEAM_WORLD)
    return team_sync_block(team)


@cute.jit
def sync_all_warp():
    """
    Executes a warp-level synchronization across all PEs in the NVSHMEM runtime
    (equivalent to ``sync_warp(TEAM_WORLD)``). All threads in the warp must call this function.

    This is a convenience wrapper around ``sync_warp`` that automatically uses
    ``TEAM_WORLD`` as the team, covering all PEs participating in the NVSHMEM job.

    Note:
        All PEs must call ``sync_all_warp`` before any PE can proceed past it.
    """
    team = _resolve_team(nvshmem.core.Teams.TEAM_WORLD)
    return team_sync_warp(team)


@cute.jit
def sync_all():
    """
    Executes a thread-level synchronization across all PEs in the NVSHMEM runtime
    (equivalent to ``sync(TEAM_WORLD)``).

    This is a convenience wrapper around ``sync`` that automatically uses
    ``TEAM_WORLD`` as the team, covering all PEs participating in the NVSHMEM job.

    Note:
        All PEs must call ``sync_all`` before any PE can proceed past it.
    """
    team = _resolve_team(nvshmem.core.Teams.TEAM_WORLD)
    return team_sync(team)


# barrier variations


@cute.jit
def barrier_block(team):
    """
    Executes a CTA-level barrier across all PEs in ``team``. All threads in the CTA must call this function.

    A barrier combines synchronization with a full memory fence, ensuring that all
    outstanding NVSHMEM memory operations (puts, gets, atomics) issued before the barrier
    are complete and visible before any PE in the team proceeds past the barrier.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the set of PEs to barrier on.
          Use ``nvshmem.core.Teams.TEAM_WORLD`` to barrier across all PEs.

    Note:
        All PEs in ``team`` must call ``barrier_block`` with the same ``team`` argument.
        Use ``barrier_all_block`` to barrier across all PEs without specifying a team.
        ``barrier`` provides stronger ordering guarantees than ``sync``.
    """
    team = _resolve_team(team)
    return _nvshmem_barrier_block(team)


@cute.jit
def barrier_warp(team):
    """
    Executes a warp-level barrier across all PEs in ``team``. All threads in the warp must call this function.

    A barrier combines synchronization with a full memory fence, ensuring that all
    outstanding NVSHMEM memory operations (puts, gets, atomics) issued before the barrier
    are complete and visible before any PE in the team proceeds past the barrier.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the set of PEs to barrier on.
          Use ``nvshmem.core.Teams.TEAM_WORLD`` to barrier across all PEs.

    Note:
        All PEs in ``team`` must call ``barrier_warp`` with the same ``team`` argument.
        Use ``barrier_all_warp`` to barrier across all PEs without specifying a team.
        ``barrier`` provides stronger ordering guarantees than ``sync``.
    """
    team = _resolve_team(team)
    return _nvshmem_barrier_warp(team)


@cute.jit
def barrier(team):
    """
    Executes a thread-level barrier across all PEs in ``team``.

    A barrier combines synchronization with a full memory fence, ensuring that all
    outstanding NVSHMEM memory operations (puts, gets, atomics) issued before the barrier
    are complete and visible before any PE in the team proceeds past the barrier.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the set of PEs to barrier on.
          Use ``nvshmem.core.Teams.TEAM_WORLD`` to barrier across all PEs.

    Note:
        All PEs in ``team`` must call ``barrier`` with the same ``team`` argument.
        Use ``barrier_all`` to barrier across all PEs without specifying a team.
        ``barrier`` provides stronger ordering guarantees than ``sync``.
    """
    team = _resolve_team(team)
    return _nvshmem_barrier(team)


# barrier_all variations


@cute.jit
def barrier_all_block():
    """
    Executes a CTA-level barrier across all PEs in the NVSHMEM runtime
    (equivalent to ``barrier_block(TEAM_WORLD)``). All threads in the CTA must call this function.

    This is a convenience wrapper around ``barrier_block`` that automatically uses
    ``TEAM_WORLD`` as the team. It combines synchronization with a full memory fence,
    ensuring all outstanding NVSHMEM memory operations are visible before proceeding.

    Note:
        All PEs must call ``barrier_all_block`` before any PE can proceed past it.
        ``barrier_all`` provides stronger ordering guarantees than ``sync_all``.
    """
    team = _resolve_team(nvshmem.core.Teams.TEAM_WORLD)
    return _nvshmem_barrier_block(team)


@cute.jit
def barrier_all_warp():
    """
    Executes a warp-level barrier across all PEs in the NVSHMEM runtime
    (equivalent to ``barrier_warp(TEAM_WORLD)``). All threads in the warp must call this function.

    This is a convenience wrapper around ``barrier_warp`` that automatically uses
    ``TEAM_WORLD`` as the team. It combines synchronization with a full memory fence,
    ensuring all outstanding NVSHMEM memory operations are visible before proceeding.

    Note:
        All PEs must call ``barrier_all_warp`` before any PE can proceed past it.
        ``barrier_all`` provides stronger ordering guarantees than ``sync_all``.
    """
    team = _resolve_team(nvshmem.core.Teams.TEAM_WORLD)
    return _nvshmem_barrier_warp(team)


@cute.jit
def barrier_all():
    """
    Executes a thread-level barrier across all PEs in the NVSHMEM runtime
    (equivalent to ``barrier(TEAM_WORLD)``).

    This is a convenience wrapper around ``barrier`` that automatically uses
    ``TEAM_WORLD`` as the team. It combines synchronization with a full memory fence,
    ensuring all outstanding NVSHMEM memory operations are visible before proceeding.

    Note:
        All PEs must call ``barrier_all`` before any PE can proceed past it.
        ``barrier_all`` provides stronger ordering guarantees than ``sync_all``.
    """
    team = _resolve_team(nvshmem.core.Teams.TEAM_WORLD)
    return _nvshmem_barrier(team)


# reduce variations


@cute.jit
def reduce_block(team, dst, src, op):
    """
    Performs a CTA-scoped all-reduce from ``src`` to ``dst`` across all PEs in ``team``. All threads in the CTA must call this function with the same arguments.

    Each PE contributes ``size(dst)`` elements from ``src``, and the result of applying
    the reduction operator ``op`` element-wise across all PEs is written to ``dst`` on
    every PE in the team (all-reduce semantics).

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array.
          Must be a symmetric (NVSHMEM-allocated) tensor. The number of elements reduced
          is ``size(dst)``.
        - ``src``: CuTe tensor view pointing to the symmetric source array.
          Must be a symmetric (NVSHMEM-allocated) tensor with at least ``size(dst)`` elements.
        - ``op`` (``str``): Reduction operator string. Supported operators for numeric types:
          ``"sum"``, ``"prod"``, ``"min"``, ``"max"``. Additional bitwise operators for
          integral types: ``"and"``, ``"or"``, ``"xor"``.

    Note:
        All PEs in ``team`` must call ``reduce_block`` before any PE can proceed past it.
        The element count is taken from ``dst``. Passing an unsupported op/dtype combination
        raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_dst(dst)
    dtype = _resolve_dtype(dst)

    if const_expr(op == "min"):

        if const_expr(dtype == cutlass.Int8):
            return int8_min_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_min_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_min_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_min_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_min_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_min_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_min_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_min_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_min_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_min_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_min_reduce_block(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "max"):

        if const_expr(dtype == cutlass.Int8):
            return int8_max_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_max_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_max_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_max_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_max_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_max_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_max_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_max_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_max_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_max_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_max_reduce_block(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "sum"):

        if const_expr(dtype == cutlass.Int8):
            return int8_sum_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_sum_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_sum_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_sum_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_sum_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_sum_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_sum_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_sum_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_sum_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_sum_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_sum_reduce_block(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "prod"):

        if const_expr(dtype == cutlass.Int8):
            return int8_prod_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_prod_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_prod_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_prod_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_prod_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_prod_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_prod_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_prod_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_prod_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_prod_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_prod_reduce_block(team, dst_ptr, src_ptr, nelem)



    elif const_expr(op == "and"):

        if const_expr(dtype == cutlass.Int8):
            return int8_and_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_and_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_and_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_and_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_and_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_and_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_and_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_and_reduce_block(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "or"):

        if const_expr(dtype == cutlass.Int8):
            return int8_or_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_or_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_or_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_or_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_or_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_or_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_or_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_or_reduce_block(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "xor"):

        if const_expr(dtype == cutlass.Int8):
            return int8_xor_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_xor_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_xor_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_xor_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_xor_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_xor_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_xor_reduce_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_xor_reduce_block(team, dst_ptr, src_ptr, nelem)


    raise RuntimeError(f"Unsupported CuTe reduce op/dtype combination: op={op}, dtype={dtype}")


@cute.jit
def reduce_warp(team, dst, src, op):
    """
    Performs a warp-scoped all-reduce from ``src`` to ``dst`` across all PEs in ``team``. All threads in the warp must call this function with the same arguments.

    Each PE contributes ``size(dst)`` elements from ``src``, and the result of applying
    the reduction operator ``op`` element-wise across all PEs is written to ``dst`` on
    every PE in the team (all-reduce semantics).

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array.
          Must be a symmetric (NVSHMEM-allocated) tensor. The number of elements reduced
          is ``size(dst)``.
        - ``src``: CuTe tensor view pointing to the symmetric source array.
          Must be a symmetric (NVSHMEM-allocated) tensor with at least ``size(dst)`` elements.
        - ``op`` (``str``): Reduction operator string. Supported operators for numeric types:
          ``"sum"``, ``"prod"``, ``"min"``, ``"max"``. Additional bitwise operators for
          integral types: ``"and"``, ``"or"``, ``"xor"``.

    Note:
        All PEs in ``team`` must call ``reduce_warp`` before any PE can proceed past it.
        The element count is taken from ``dst``. Passing an unsupported op/dtype combination
        raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_dst(dst)
    dtype = _resolve_dtype(dst)

    if const_expr(op == "min"):

        if const_expr(dtype == cutlass.Int8):
            return int8_min_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_min_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_min_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_min_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_min_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_min_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_min_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_min_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_min_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_min_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_min_reduce_warp(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "max"):

        if const_expr(dtype == cutlass.Int8):
            return int8_max_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_max_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_max_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_max_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_max_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_max_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_max_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_max_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_max_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_max_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_max_reduce_warp(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "sum"):

        if const_expr(dtype == cutlass.Int8):
            return int8_sum_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_sum_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_sum_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_sum_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_sum_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_sum_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_sum_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_sum_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_sum_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_sum_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_sum_reduce_warp(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "prod"):

        if const_expr(dtype == cutlass.Int8):
            return int8_prod_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_prod_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_prod_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_prod_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_prod_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_prod_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_prod_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_prod_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_prod_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_prod_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_prod_reduce_warp(team, dst_ptr, src_ptr, nelem)



    elif const_expr(op == "and"):

        if const_expr(dtype == cutlass.Int8):
            return int8_and_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_and_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_and_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_and_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_and_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_and_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_and_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_and_reduce_warp(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "or"):

        if const_expr(dtype == cutlass.Int8):
            return int8_or_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_or_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_or_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_or_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_or_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_or_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_or_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_or_reduce_warp(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "xor"):

        if const_expr(dtype == cutlass.Int8):
            return int8_xor_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_xor_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_xor_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_xor_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_xor_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_xor_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_xor_reduce_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_xor_reduce_warp(team, dst_ptr, src_ptr, nelem)


    raise RuntimeError(f"Unsupported CuTe reduce op/dtype combination: op={op}, dtype={dtype}")


@cute.jit
def reduce(team, dst, src, op):
    """
    Performs a thread-scoped all-reduce from ``src`` to ``dst`` across all PEs in ``team``.

    Each PE contributes ``size(dst)`` elements from ``src``, and the result of applying
    the reduction operator ``op`` element-wise across all PEs is written to ``dst`` on
    every PE in the team (all-reduce semantics).

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array.
          Must be a symmetric (NVSHMEM-allocated) tensor. The number of elements reduced
          is ``size(dst)``.
        - ``src``: CuTe tensor view pointing to the symmetric source array.
          Must be a symmetric (NVSHMEM-allocated) tensor with at least ``size(dst)`` elements.
        - ``op`` (``str``): Reduction operator string. Supported operators for numeric types:
          ``"sum"``, ``"prod"``, ``"min"``, ``"max"``. Additional bitwise operators for
          integral types: ``"and"``, ``"or"``, ``"xor"``.

    Note:
        All PEs in ``team`` must call ``reduce`` before any PE can proceed past it.
        The element count is taken from ``dst``. Passing an unsupported op/dtype combination
        raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_dst(dst)
    dtype = _resolve_dtype(dst)

    if const_expr(op == "min"):

        if const_expr(dtype == cutlass.Int8):
            return int8_min_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_min_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_min_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_min_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_min_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_min_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_min_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_min_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_min_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_min_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_min_reduce(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "max"):

        if const_expr(dtype == cutlass.Int8):
            return int8_max_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_max_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_max_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_max_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_max_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_max_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_max_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_max_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_max_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_max_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_max_reduce(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "sum"):

        if const_expr(dtype == cutlass.Int8):
            return int8_sum_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_sum_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_sum_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_sum_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_sum_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_sum_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_sum_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_sum_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_sum_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_sum_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_sum_reduce(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "prod"):

        if const_expr(dtype == cutlass.Int8):
            return int8_prod_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_prod_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_prod_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_prod_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_prod_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_prod_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_prod_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_prod_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_prod_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_prod_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_prod_reduce(team, dst_ptr, src_ptr, nelem)



    elif const_expr(op == "and"):

        if const_expr(dtype == cutlass.Int8):
            return int8_and_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_and_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_and_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_and_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_and_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_and_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_and_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_and_reduce(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "or"):

        if const_expr(dtype == cutlass.Int8):
            return int8_or_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_or_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_or_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_or_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_or_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_or_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_or_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_or_reduce(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "xor"):

        if const_expr(dtype == cutlass.Int8):
            return int8_xor_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_xor_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_xor_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_xor_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_xor_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_xor_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_xor_reduce(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_xor_reduce(team, dst_ptr, src_ptr, nelem)


    raise RuntimeError(f"Unsupported CuTe reduce op/dtype combination: op={op}, dtype={dtype}")


# reducescatter variations


@cute.jit
def reducescatter_block(team, dst, src, op):
    """
    Performs a CTA-scoped reduce-scatter from ``src`` to ``dst`` across all PEs in ``team``. All threads in the CTA must call this function with the same arguments.

    In a reduce-scatter, each PE contributes elements from ``src``, and the result of
    applying the reduction operator element-wise across all PEs is divided into equal
    portions, each portion written to ``dst`` on a different PE (scatter semantics).

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array.
          Must be a symmetric (NVSHMEM-allocated) tensor. The number of output elements
          per PE is ``size(dst)``. The ``src`` array should have at least
          ``size(dst) * team_n_pes(team)`` elements.
        - ``src``: CuTe tensor view pointing to the symmetric source array.
          Must be a symmetric (NVSHMEM-allocated) tensor.
        - ``op`` (``str``): Reduction operator string. Supported operators for numeric types:
          ``"sum"``, ``"prod"``, ``"min"``, ``"max"``. Additional bitwise operators for
          integral types: ``"and"``, ``"or"``, ``"xor"``.

    Note:
        All PEs in ``team`` must call ``reducescatter_block`` before any PE can proceed past it.
        The element count is taken from ``dst``. Passing an unsupported op/dtype combination
        raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_dst(dst)
    dtype = _resolve_dtype(dst)

    if const_expr(op == "min"):

        if const_expr(dtype == cutlass.Int8):
            return int8_min_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_min_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_min_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_min_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_min_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_min_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_min_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_min_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_min_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_min_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_min_reducescatter_block(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "max"):

        if const_expr(dtype == cutlass.Int8):
            return int8_max_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_max_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_max_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_max_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_max_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_max_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_max_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_max_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_max_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_max_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_max_reducescatter_block(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "sum"):

        if const_expr(dtype == cutlass.Int8):
            return int8_sum_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_sum_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_sum_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_sum_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_sum_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_sum_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_sum_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_sum_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_sum_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_sum_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_sum_reducescatter_block(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "prod"):

        if const_expr(dtype == cutlass.Int8):
            return int8_prod_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_prod_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_prod_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_prod_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_prod_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_prod_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_prod_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_prod_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_prod_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_prod_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_prod_reducescatter_block(team, dst_ptr, src_ptr, nelem)



    elif const_expr(op == "and"):

        if const_expr(dtype == cutlass.Int8):
            return int8_and_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_and_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_and_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_and_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_and_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_and_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_and_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_and_reducescatter_block(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "or"):

        if const_expr(dtype == cutlass.Int8):
            return int8_or_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_or_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_or_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_or_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_or_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_or_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_or_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_or_reducescatter_block(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "xor"):

        if const_expr(dtype == cutlass.Int8):
            return int8_xor_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_xor_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_xor_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_xor_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_xor_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_xor_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_xor_reducescatter_block(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_xor_reducescatter_block(team, dst_ptr, src_ptr, nelem)


    raise RuntimeError(f"Unsupported CuTe reducescatter op/dtype combination: op={op}, dtype={dtype}")


@cute.jit
def reducescatter_warp(team, dst, src, op):
    """
    Performs a warp-scoped reduce-scatter from ``src`` to ``dst`` across all PEs in ``team``. All threads in the warp must call this function with the same arguments.

    In a reduce-scatter, each PE contributes elements from ``src``, and the result of
    applying the reduction operator element-wise across all PEs is divided into equal
    portions, each portion written to ``dst`` on a different PE (scatter semantics).

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array.
          Must be a symmetric (NVSHMEM-allocated) tensor. The number of output elements
          per PE is ``size(dst)``. The ``src`` array should have at least
          ``size(dst) * team_n_pes(team)`` elements.
        - ``src``: CuTe tensor view pointing to the symmetric source array.
          Must be a symmetric (NVSHMEM-allocated) tensor.
        - ``op`` (``str``): Reduction operator string. Supported operators for numeric types:
          ``"sum"``, ``"prod"``, ``"min"``, ``"max"``. Additional bitwise operators for
          integral types: ``"and"``, ``"or"``, ``"xor"``.

    Note:
        All PEs in ``team`` must call ``reducescatter_warp`` before any PE can proceed past it.
        The element count is taken from ``dst``. Passing an unsupported op/dtype combination
        raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_dst(dst)
    dtype = _resolve_dtype(dst)

    if const_expr(op == "min"):

        if const_expr(dtype == cutlass.Int8):
            return int8_min_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_min_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_min_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_min_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_min_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_min_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_min_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_min_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_min_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_min_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_min_reducescatter_warp(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "max"):

        if const_expr(dtype == cutlass.Int8):
            return int8_max_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_max_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_max_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_max_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_max_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_max_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_max_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_max_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_max_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_max_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_max_reducescatter_warp(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "sum"):

        if const_expr(dtype == cutlass.Int8):
            return int8_sum_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_sum_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_sum_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_sum_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_sum_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_sum_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_sum_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_sum_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_sum_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_sum_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_sum_reducescatter_warp(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "prod"):

        if const_expr(dtype == cutlass.Int8):
            return int8_prod_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_prod_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_prod_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_prod_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_prod_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_prod_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_prod_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_prod_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_prod_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_prod_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_prod_reducescatter_warp(team, dst_ptr, src_ptr, nelem)



    elif const_expr(op == "and"):

        if const_expr(dtype == cutlass.Int8):
            return int8_and_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_and_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_and_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_and_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_and_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_and_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_and_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_and_reducescatter_warp(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "or"):

        if const_expr(dtype == cutlass.Int8):
            return int8_or_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_or_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_or_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_or_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_or_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_or_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_or_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_or_reducescatter_warp(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "xor"):

        if const_expr(dtype == cutlass.Int8):
            return int8_xor_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_xor_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_xor_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_xor_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_xor_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_xor_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_xor_reducescatter_warp(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_xor_reducescatter_warp(team, dst_ptr, src_ptr, nelem)


    raise RuntimeError(f"Unsupported CuTe reducescatter op/dtype combination: op={op}, dtype={dtype}")


@cute.jit
def reducescatter(team, dst, src, op):
    """
    Performs a thread-scoped reduce-scatter from ``src`` to ``dst`` across all PEs in ``team``.

    In a reduce-scatter, each PE contributes elements from ``src``, and the result of
    applying the reduction operator element-wise across all PEs is divided into equal
    portions, each portion written to ``dst`` on a different PE (scatter semantics).

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array.
          Must be a symmetric (NVSHMEM-allocated) tensor. The number of output elements
          per PE is ``size(dst)``. The ``src`` array should have at least
          ``size(dst) * team_n_pes(team)`` elements.
        - ``src``: CuTe tensor view pointing to the symmetric source array.
          Must be a symmetric (NVSHMEM-allocated) tensor.
        - ``op`` (``str``): Reduction operator string. Supported operators for numeric types:
          ``"sum"``, ``"prod"``, ``"min"``, ``"max"``. Additional bitwise operators for
          integral types: ``"and"``, ``"or"``, ``"xor"``.

    Note:
        All PEs in ``team`` must call ``reducescatter`` before any PE can proceed past it.
        The element count is taken from ``dst``. Passing an unsupported op/dtype combination
        raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_dst(dst)
    dtype = _resolve_dtype(dst)

    if const_expr(op == "min"):

        if const_expr(dtype == cutlass.Int8):
            return int8_min_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_min_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_min_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_min_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_min_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_min_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_min_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_min_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_min_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_min_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_min_reducescatter(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "max"):

        if const_expr(dtype == cutlass.Int8):
            return int8_max_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_max_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_max_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_max_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_max_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_max_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_max_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_max_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_max_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_max_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_max_reducescatter(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "sum"):

        if const_expr(dtype == cutlass.Int8):
            return int8_sum_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_sum_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_sum_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_sum_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_sum_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_sum_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_sum_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_sum_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_sum_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_sum_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_sum_reducescatter(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "prod"):

        if const_expr(dtype == cutlass.Int8):
            return int8_prod_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_prod_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_prod_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_prod_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_prod_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_prod_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_prod_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_prod_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float32):
            return float_prod_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float64):
            return double_prod_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Float16):
            return half_prod_reducescatter(team, dst_ptr, src_ptr, nelem)



    elif const_expr(op == "and"):

        if const_expr(dtype == cutlass.Int8):
            return int8_and_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_and_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_and_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_and_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_and_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_and_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_and_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_and_reducescatter(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "or"):

        if const_expr(dtype == cutlass.Int8):
            return int8_or_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_or_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_or_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_or_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_or_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_or_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_or_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_or_reducescatter(team, dst_ptr, src_ptr, nelem)


    elif const_expr(op == "xor"):

        if const_expr(dtype == cutlass.Int8):
            return int8_xor_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int16):
            return int16_xor_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int32):
            return int32_xor_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Int64):
            return int64_xor_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint8):
            return uint8_xor_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint16):
            return uint16_xor_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint32):
            return uint32_xor_reducescatter(team, dst_ptr, src_ptr, nelem)

        elif const_expr(dtype == cutlass.Uint64):
            return uint64_xor_reducescatter(team, dst_ptr, src_ptr, nelem)


    raise RuntimeError(f"Unsupported CuTe reducescatter op/dtype combination: op={op}, dtype={dtype}")


# fcollect variations


@cute.jit
def fcollect_block(team, dst, src):
    """
    Performs a CTA-scoped fcollect (all-gather) from ``src`` to ``dst`` across all PEs in ``team``. All threads in the CTA must call this function with the same arguments.

    Each PE contributes ``size(src)`` elements, and the concatenated result from all PEs
    is written to ``dst`` on every PE. The ``dst`` array must be large enough to hold
    contributions from all PEs: ``size(dst) >= size(src) * team_n_pes(team)``.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array on this PE.
          Must be a symmetric (NVSHMEM-allocated) tensor with at least
          ``size(src) * team_n_pes(team)`` elements.
        - ``src``: CuTe tensor view pointing to the symmetric source array on this PE.
          Must be a symmetric (NVSHMEM-allocated) tensor. The number of elements
          contributed per PE is ``size(src)``.

    Note:
        All PEs in ``team`` must call ``fcollect_block`` before any PE can proceed past it.
        The element count per PE is taken from ``src``.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_src(src)
    dtype = _resolve_dtype(dst)

    if const_expr(dtype == cutlass.Int8):
        return int8_fcollect_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int16):
        return int16_fcollect_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int32):
        return int32_fcollect_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int64):
        return int64_fcollect_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint8):
        return uint8_fcollect_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint16):
        return uint16_fcollect_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint32):
        return uint32_fcollect_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint64):
        return uint64_fcollect_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float32):
        return float_fcollect_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float64):
        return double_fcollect_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float16):
        return half_fcollect_block(team, dst_ptr, src_ptr, nelem)

    raise RuntimeError(f"Unsupported CuTe dtype for fcollect: {dtype}")


@cute.jit
def fcollect_warp(team, dst, src):
    """
    Performs a warp-scoped fcollect (all-gather) from ``src`` to ``dst`` across all PEs in ``team``. All threads in the warp must call this function with the same arguments.

    Each PE contributes ``size(src)`` elements, and the concatenated result from all PEs
    is written to ``dst`` on every PE. The ``dst`` array must be large enough to hold
    contributions from all PEs: ``size(dst) >= size(src) * team_n_pes(team)``.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array on this PE.
          Must be a symmetric (NVSHMEM-allocated) tensor with at least
          ``size(src) * team_n_pes(team)`` elements.
        - ``src``: CuTe tensor view pointing to the symmetric source array on this PE.
          Must be a symmetric (NVSHMEM-allocated) tensor. The number of elements
          contributed per PE is ``size(src)``.

    Note:
        All PEs in ``team`` must call ``fcollect_warp`` before any PE can proceed past it.
        The element count per PE is taken from ``src``.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_src(src)
    dtype = _resolve_dtype(dst)

    if const_expr(dtype == cutlass.Int8):
        return int8_fcollect_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int16):
        return int16_fcollect_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int32):
        return int32_fcollect_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int64):
        return int64_fcollect_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint8):
        return uint8_fcollect_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint16):
        return uint16_fcollect_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint32):
        return uint32_fcollect_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint64):
        return uint64_fcollect_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float32):
        return float_fcollect_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float64):
        return double_fcollect_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float16):
        return half_fcollect_warp(team, dst_ptr, src_ptr, nelem)

    raise RuntimeError(f"Unsupported CuTe dtype for fcollect: {dtype}")


@cute.jit
def fcollect(team, dst, src):
    """
    Performs a thread-scoped fcollect (all-gather) from ``src`` to ``dst`` across all PEs in ``team``.

    Each PE contributes ``size(src)`` elements, and the concatenated result from all PEs
    is written to ``dst`` on every PE. The ``dst`` array must be large enough to hold
    contributions from all PEs: ``size(dst) >= size(src) * team_n_pes(team)``.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array on this PE.
          Must be a symmetric (NVSHMEM-allocated) tensor with at least
          ``size(src) * team_n_pes(team)`` elements.
        - ``src``: CuTe tensor view pointing to the symmetric source array on this PE.
          Must be a symmetric (NVSHMEM-allocated) tensor. The number of elements
          contributed per PE is ``size(src)``.

    Note:
        All PEs in ``team`` must call ``fcollect`` before any PE can proceed past it.
        The element count per PE is taken from ``src``.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_src(src)
    dtype = _resolve_dtype(dst)

    if const_expr(dtype == cutlass.Int8):
        return int8_fcollect(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int16):
        return int16_fcollect(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int32):
        return int32_fcollect(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int64):
        return int64_fcollect(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint8):
        return uint8_fcollect(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint16):
        return uint16_fcollect(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint32):
        return uint32_fcollect(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint64):
        return uint64_fcollect(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float32):
        return float_fcollect(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float64):
        return double_fcollect(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float16):
        return half_fcollect(team, dst_ptr, src_ptr, nelem)

    raise RuntimeError(f"Unsupported CuTe dtype for fcollect: {dtype}")


# broadcast variations


@cute.jit
def broadcast_block(team, dst, src, root=0):
    """
    Performs a CTA-scoped broadcast from ``src`` on ``root`` PE to ``dst`` on all PEs in ``team``. All threads in the CTA must call this function with the same arguments.

    The root PE broadcasts the contents of its ``src`` array to the ``dst`` array on
    every PE in the team (including the root itself). Non-root PEs ignore their own ``src``.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array.
          Must be a symmetric (NVSHMEM-allocated) tensor on all PEs.
        - ``src``: CuTe tensor view pointing to the symmetric source array on the root PE.
          Must be a symmetric (NVSHMEM-allocated) tensor. Only the root PE's ``src`` is used.
        - ``root`` (``int``, optional): PE rank within ``team`` that serves as the broadcast source.
          Defaults to ``0`` (the first PE in the team).

    Note:
        All PEs in ``team`` must call ``broadcast_block`` before any PE can proceed past it.
        The number of elements broadcast is ``min(size(dst), size(src))``.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_min(dst, src)
    root = _resolve_root(root)
    dtype = _resolve_dtype(dst)

    if const_expr(dtype == cutlass.Int8):
        return int8_broadcast_block(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Int16):
        return int16_broadcast_block(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Int32):
        return int32_broadcast_block(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Int64):
        return int64_broadcast_block(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Uint8):
        return uint8_broadcast_block(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Uint16):
        return uint16_broadcast_block(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Uint32):
        return uint32_broadcast_block(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Uint64):
        return uint64_broadcast_block(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Float32):
        return float_broadcast_block(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Float64):
        return double_broadcast_block(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Float16):
        return half_broadcast_block(team, dst_ptr, src_ptr, nelem, root)

    raise RuntimeError(f"Unsupported CuTe dtype for broadcast: {dtype}")


@cute.jit
def broadcast_warp(team, dst, src, root=0):
    """
    Performs a warp-scoped broadcast from ``src`` on ``root`` PE to ``dst`` on all PEs in ``team``. All threads in the warp must call this function with the same arguments.

    The root PE broadcasts the contents of its ``src`` array to the ``dst`` array on
    every PE in the team (including the root itself). Non-root PEs ignore their own ``src``.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array.
          Must be a symmetric (NVSHMEM-allocated) tensor on all PEs.
        - ``src``: CuTe tensor view pointing to the symmetric source array on the root PE.
          Must be a symmetric (NVSHMEM-allocated) tensor. Only the root PE's ``src`` is used.
        - ``root`` (``int``, optional): PE rank within ``team`` that serves as the broadcast source.
          Defaults to ``0`` (the first PE in the team).

    Note:
        All PEs in ``team`` must call ``broadcast_warp`` before any PE can proceed past it.
        The number of elements broadcast is ``min(size(dst), size(src))``.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_min(dst, src)
    root = _resolve_root(root)
    dtype = _resolve_dtype(dst)

    if const_expr(dtype == cutlass.Int8):
        return int8_broadcast_warp(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Int16):
        return int16_broadcast_warp(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Int32):
        return int32_broadcast_warp(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Int64):
        return int64_broadcast_warp(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Uint8):
        return uint8_broadcast_warp(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Uint16):
        return uint16_broadcast_warp(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Uint32):
        return uint32_broadcast_warp(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Uint64):
        return uint64_broadcast_warp(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Float32):
        return float_broadcast_warp(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Float64):
        return double_broadcast_warp(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Float16):
        return half_broadcast_warp(team, dst_ptr, src_ptr, nelem, root)

    raise RuntimeError(f"Unsupported CuTe dtype for broadcast: {dtype}")


@cute.jit
def broadcast(team, dst, src, root=0):
    """
    Performs a thread-scoped broadcast from ``src`` on ``root`` PE to ``dst`` on all PEs in ``team``.

    The root PE broadcasts the contents of its ``src`` array to the ``dst`` array on
    every PE in the team (including the root itself). Non-root PEs ignore their own ``src``.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array.
          Must be a symmetric (NVSHMEM-allocated) tensor on all PEs.
        - ``src``: CuTe tensor view pointing to the symmetric source array on the root PE.
          Must be a symmetric (NVSHMEM-allocated) tensor. Only the root PE's ``src`` is used.
        - ``root`` (``int``, optional): PE rank within ``team`` that serves as the broadcast source.
          Defaults to ``0`` (the first PE in the team).

    Note:
        All PEs in ``team`` must call ``broadcast`` before any PE can proceed past it.
        The number of elements broadcast is ``min(size(dst), size(src))``.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_min(dst, src)
    root = _resolve_root(root)
    dtype = _resolve_dtype(dst)

    if const_expr(dtype == cutlass.Int8):
        return int8_broadcast(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Int16):
        return int16_broadcast(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Int32):
        return int32_broadcast(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Int64):
        return int64_broadcast(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Uint8):
        return uint8_broadcast(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Uint16):
        return uint16_broadcast(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Uint32):
        return uint32_broadcast(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Uint64):
        return uint64_broadcast(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Float32):
        return float_broadcast(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Float64):
        return double_broadcast(team, dst_ptr, src_ptr, nelem, root)

    elif const_expr(dtype == cutlass.Float16):
        return half_broadcast(team, dst_ptr, src_ptr, nelem, root)

    raise RuntimeError(f"Unsupported CuTe dtype for broadcast: {dtype}")


# alltoall variations


@cute.jit
def alltoall_block(team, dst, src):
    """
    Performs a CTA-scoped all-to-all exchange from ``src`` to ``dst`` across all PEs in ``team``. All threads in the CTA must call this function with the same arguments.

    In an all-to-all operation, each PE sends a distinct portion of its ``src`` array to
    every other PE, and receives a portion from every PE into its ``dst`` array.
    The ``src`` array is logically divided into ``team_n_pes(team)`` equal segments;
    segment ``i`` is sent to PE ``i``. Each PE receives one segment from every PE into
    the corresponding portion of ``dst``.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array on this PE.
          Must be a symmetric (NVSHMEM-allocated) tensor. Must have at least
          ``size(src)`` elements (same total size as ``src``).
        - ``src``: CuTe tensor view pointing to the symmetric source array on this PE.
          Must be a symmetric (NVSHMEM-allocated) tensor. The per-PE send count is
          ``size(src) // team_n_pes(team)`` elements.

    Note:
        All PEs in ``team`` must call ``alltoall_block`` before any PE can proceed past it.
        ``size(src)`` must be evenly divisible by ``team_n_pes(team)``.
        The per-PE element count passed to NVSHMEM is ``size(src) // team_n_pes(team)``.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_alltoall(src, team)
    dtype = _resolve_dtype(dst)

    if const_expr(dtype == cutlass.Int8):
        return int8_alltoall_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int16):
        return int16_alltoall_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int32):
        return int32_alltoall_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int64):
        return int64_alltoall_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint8):
        return uint8_alltoall_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint16):
        return uint16_alltoall_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint32):
        return uint32_alltoall_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint64):
        return uint64_alltoall_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float32):
        return float_alltoall_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float64):
        return double_alltoall_block(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float16):
        return half_alltoall_block(team, dst_ptr, src_ptr, nelem)

    raise RuntimeError(f"Unsupported CuTe dtype for alltoall: {dtype}")


@cute.jit
def alltoall_warp(team, dst, src):
    """
    Performs a warp-scoped all-to-all exchange from ``src`` to ``dst`` across all PEs in ``team``. All threads in the warp must call this function with the same arguments.

    In an all-to-all operation, each PE sends a distinct portion of its ``src`` array to
    every other PE, and receives a portion from every PE into its ``dst`` array.
    The ``src`` array is logically divided into ``team_n_pes(team)`` equal segments;
    segment ``i`` is sent to PE ``i``. Each PE receives one segment from every PE into
    the corresponding portion of ``dst``.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array on this PE.
          Must be a symmetric (NVSHMEM-allocated) tensor. Must have at least
          ``size(src)`` elements (same total size as ``src``).
        - ``src``: CuTe tensor view pointing to the symmetric source array on this PE.
          Must be a symmetric (NVSHMEM-allocated) tensor. The per-PE send count is
          ``size(src) // team_n_pes(team)`` elements.

    Note:
        All PEs in ``team`` must call ``alltoall_warp`` before any PE can proceed past it.
        ``size(src)`` must be evenly divisible by ``team_n_pes(team)``.
        The per-PE element count passed to NVSHMEM is ``size(src) // team_n_pes(team)``.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_alltoall(src, team)
    dtype = _resolve_dtype(dst)

    if const_expr(dtype == cutlass.Int8):
        return int8_alltoall_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int16):
        return int16_alltoall_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int32):
        return int32_alltoall_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int64):
        return int64_alltoall_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint8):
        return uint8_alltoall_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint16):
        return uint16_alltoall_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint32):
        return uint32_alltoall_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint64):
        return uint64_alltoall_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float32):
        return float_alltoall_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float64):
        return double_alltoall_warp(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float16):
        return half_alltoall_warp(team, dst_ptr, src_ptr, nelem)

    raise RuntimeError(f"Unsupported CuTe dtype for alltoall: {dtype}")


@cute.jit
def alltoall(team, dst, src):
    """
    Performs a thread-scoped all-to-all exchange from ``src`` to ``dst`` across all PEs in ``team``.

    In an all-to-all operation, each PE sends a distinct portion of its ``src`` array to
    every other PE, and receives a portion from every PE into its ``dst`` array.
    The ``src`` array is logically divided into ``team_n_pes(team)`` equal segments;
    segment ``i`` is sent to PE ``i``. Each PE receives one segment from every PE into
    the corresponding portion of ``dst``.

    Args:
        - ``team`` (``int``): NVSHMEM team handle identifying the participating PEs.
        - ``dst``: CuTe tensor view pointing to the symmetric destination array on this PE.
          Must be a symmetric (NVSHMEM-allocated) tensor. Must have at least
          ``size(src)`` elements (same total size as ``src``).
        - ``src``: CuTe tensor view pointing to the symmetric source array on this PE.
          Must be a symmetric (NVSHMEM-allocated) tensor. The per-PE send count is
          ``size(src) // team_n_pes(team)`` elements.

    Note:
        All PEs in ``team`` must call ``alltoall`` before any PE can proceed past it.
        ``size(src)`` must be evenly divisible by ``team_n_pes(team)``.
        The per-PE element count passed to NVSHMEM is ``size(src) // team_n_pes(team)``.
        Passing an unsupported dtype raises ``RuntimeError`` at JIT compile time.
    """
    team = _resolve_team(team)
    dst_ptr = _resolve_ptr(dst)
    src_ptr = _resolve_ptr(src)
    nelem = _resolve_nelems_alltoall(src, team)
    dtype = _resolve_dtype(dst)

    if const_expr(dtype == cutlass.Int8):
        return int8_alltoall(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int16):
        return int16_alltoall(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int32):
        return int32_alltoall(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Int64):
        return int64_alltoall(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint8):
        return uint8_alltoall(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint16):
        return uint16_alltoall(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint32):
        return uint32_alltoall(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Uint64):
        return uint64_alltoall(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float32):
        return float_alltoall(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float64):
        return double_alltoall(team, dst_ptr, src_ptr, nelem)

    elif const_expr(dtype == cutlass.Float16):
        return half_alltoall(team, dst_ptr, src_ptr, nelem)

    raise RuntimeError(f"Unsupported CuTe dtype for alltoall: {dtype}")



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from nvshmem.bindings.device import numba as bindings
from nvshmem.core import Teams

import numpy as np
from numba.core.extending import overload
from numba.types import (
    int8, int16, int32, int64, uint8, uint16, uint32, uint64,
    float32, float64, float16, Array,
)
from numba.cuda.types import bfloat16
from numba.core import types, cgutils
from numba.cuda.extending import intrinsic
from numba.np import arrayobj
from numba.cuda.np.numpy_support import carray
from numba.cuda import declare_device
from numba.types import CPointer, void
from numba.cuda.cudaimpl import lower


# Imports for binding shims
from numba.cuda import CUSource
from numba.cuda.typing.templates import ConcreteTemplate
import numba
from numba.types import uint16
from numba.cuda import declare_device
from numba.cuda.cudaimpl import lower_constant
from numba.types import float32
from numba.cuda.cudadecl import register_global
from numba import types
from numba.types import int16
from numba.types import float64
from numba.types import CPointer
from numba.core.typing import signature
from numba.cuda.cudadecl import register_attr
from numba.core.extending import lower_cast
from numba.types import uint64
from numba.cuda.types import bfloat16
from numba.cuda.vector_types import vector_types
from numba.cuda.cudaimpl import lower
import io
from numba.cuda.cudadecl import register
from numba.cuda.cudaimpl import lower_attr
from numba.types import int64
from numba.types import uint32
from numba.types import float16
from numba.types import int32
from numba.types import uint8
from numba.types import bool_
from numba.types import void
from numba.types import int8

__all__ = ["get_multicast_array", "get_peer_array"]

@intrinsic
def _array_data_noneptr(typingctx, arrty):
    """
    Return the data pointer of a Numba Array as an opaque none* (CPointer(none)).
    numba-CUDA 0.21+: arr.data is T**; one load yields T*. We then bitcast to none*.
    """
    if not isinstance(arrty, types.Array):
        return None
    sig = types.CPointer(types.uint8)(arrty)

    def codegen(cgctx, builder, sig, args):
        arr_val = args[0]
        ary = arrayobj.make_array(arrty)(cgctx, builder, arr_val)
        # Bitcast T* -> i8* (opaque)
        optr = builder.bitcast(ary.data, cgutils.voidptr_t)
        return optr

    return sig, codegen

@intrinsic
def _noneptr_as_typedptr(typingctx, optr, arrty):
    """
    Bitcast an opaque none* (void*) back to T* using arr.dtype.
    """
    if not (isinstance(optr, types.CPointer) and optr.dtype is types.none):
        return None
    if not isinstance(arrty, types.Array):
        return None

    ret = types.CPointer(arrty.dtype)
    sig = ret(optr, arrty)

    def codegen(cgctx, builder, sig, args):
        optr_val, _ = args
        ret_llty = cgctx.get_value_type(sig.return_type)  # T* type
        return builder.bitcast(optr_val, ret_llty)        # T*

    return sig, codegen

# -----------------------------------------------------------------------------
# Multicast Array
# -----------------------------------------------------------------------------
def get_multicast_array(team: Teams, array: Array):
    """
    Returns an array view on multicast-accessible memory corresponding to the input array.
    The Array passed into it must be allocated by NVSHMEM4Py.

    This is the Python array equivalent of `nvshmemx_mc_ptr` which returns a pointer into
    a peer's symmetric heap

    Args:
        array: Array - A symmetric array allocated by NVSHMEM
        team: Teams - A NVSHMEM Team to create the Multicast object across

    Returns:
        - A Numba ArrayView which represents the Multicast object

    NOTE: This function is only supported with the Numba Runtime. To enable it,
          set the environment variable ``NUMBA_CUDA_ENABLE_NRT=1``
    """
    pass


@overload(get_multicast_array)
def get_multicast_array_ol(team: Teams, arr: Array) -> Array:
    
    if arr == Array(dtype=int8, ndim=arr.ndim, layout=arr.layout):
        def impl(team: Teams, arr: Array) -> Array:
            base_optr = _array_data_noneptr(arr)         # i8*
            team_int32 = np.int32(team) 
            mc_optr   = mc_ptr(team_int32, base_optr)
            mc_tptr   = _noneptr_as_typedptr(mc_optr, arr)  # T*
            return carray(mc_tptr, arr.shape, dtype=int8)
        return impl
    
    elif arr == Array(dtype=int16, ndim=arr.ndim, layout=arr.layout):
        def impl(team: Teams, arr: Array) -> Array:
            base_optr = _array_data_noneptr(arr)         # i8*
            team_int32 = np.int32(team) 
            mc_optr   = mc_ptr(team_int32, base_optr)
            mc_tptr   = _noneptr_as_typedptr(mc_optr, arr)  # T*
            return carray(mc_tptr, arr.shape, dtype=int16)
        return impl
    
    elif arr == Array(dtype=int32, ndim=arr.ndim, layout=arr.layout):
        def impl(team: Teams, arr: Array) -> Array:
            base_optr = _array_data_noneptr(arr)         # i8*
            team_int32 = np.int32(team) 
            mc_optr   = mc_ptr(team_int32, base_optr)
            mc_tptr   = _noneptr_as_typedptr(mc_optr, arr)  # T*
            return carray(mc_tptr, arr.shape, dtype=int32)
        return impl
    
    elif arr == Array(dtype=int64, ndim=arr.ndim, layout=arr.layout):
        def impl(team: Teams, arr: Array) -> Array:
            base_optr = _array_data_noneptr(arr)         # i8*
            team_int32 = np.int32(team) 
            mc_optr   = mc_ptr(team_int32, base_optr)
            mc_tptr   = _noneptr_as_typedptr(mc_optr, arr)  # T*
            return carray(mc_tptr, arr.shape, dtype=int64)
        return impl
    
    elif arr == Array(dtype=uint8, ndim=arr.ndim, layout=arr.layout):
        def impl(team: Teams, arr: Array) -> Array:
            base_optr = _array_data_noneptr(arr)         # i8*
            team_int32 = np.int32(team) 
            mc_optr   = mc_ptr(team_int32, base_optr)
            mc_tptr   = _noneptr_as_typedptr(mc_optr, arr)  # T*
            return carray(mc_tptr, arr.shape, dtype=uint8)
        return impl
    
    elif arr == Array(dtype=uint16, ndim=arr.ndim, layout=arr.layout):
        def impl(team: Teams, arr: Array) -> Array:
            base_optr = _array_data_noneptr(arr)         # i8*
            team_int32 = np.int32(team) 
            mc_optr   = mc_ptr(team_int32, base_optr)
            mc_tptr   = _noneptr_as_typedptr(mc_optr, arr)  # T*
            return carray(mc_tptr, arr.shape, dtype=uint16)
        return impl
    
    elif arr == Array(dtype=uint32, ndim=arr.ndim, layout=arr.layout):
        def impl(team: Teams, arr: Array) -> Array:
            base_optr = _array_data_noneptr(arr)         # i8*
            team_int32 = np.int32(team) 
            mc_optr   = mc_ptr(team_int32, base_optr)
            mc_tptr   = _noneptr_as_typedptr(mc_optr, arr)  # T*
            return carray(mc_tptr, arr.shape, dtype=uint32)
        return impl
    
    elif arr == Array(dtype=uint64, ndim=arr.ndim, layout=arr.layout):
        def impl(team: Teams, arr: Array) -> Array:
            base_optr = _array_data_noneptr(arr)         # i8*
            team_int32 = np.int32(team) 
            mc_optr   = mc_ptr(team_int32, base_optr)
            mc_tptr   = _noneptr_as_typedptr(mc_optr, arr)  # T*
            return carray(mc_tptr, arr.shape, dtype=uint64)
        return impl
    
    elif arr == Array(dtype=float32, ndim=arr.ndim, layout=arr.layout):
        def impl(team: Teams, arr: Array) -> Array:
            base_optr = _array_data_noneptr(arr)         # i8*
            team_int32 = np.int32(team) 
            mc_optr   = mc_ptr(team_int32, base_optr)
            mc_tptr   = _noneptr_as_typedptr(mc_optr, arr)  # T*
            return carray(mc_tptr, arr.shape, dtype=float32)
        return impl
    
    elif arr == Array(dtype=float64, ndim=arr.ndim, layout=arr.layout):
        def impl(team: Teams, arr: Array) -> Array:
            base_optr = _array_data_noneptr(arr)         # i8*
            team_int32 = np.int32(team) 
            mc_optr   = mc_ptr(team_int32, base_optr)
            mc_tptr   = _noneptr_as_typedptr(mc_optr, arr)  # T*
            return carray(mc_tptr, arr.shape, dtype=float64)
        return impl
    
    elif arr == Array(dtype=float16, ndim=arr.ndim, layout=arr.layout):
        def impl(team: Teams, arr: Array) -> Array:
            base_optr = _array_data_noneptr(arr)         # i8*
            team_int32 = np.int32(team) 
            mc_optr   = mc_ptr(team_int32, base_optr)
            mc_tptr   = _noneptr_as_typedptr(mc_optr, arr)  # T*
            return carray(mc_tptr, arr.shape, dtype=float16)
        return impl
    
    return None


# -----------------------------------------------------------------------------
# Peer Array
# -----------------------------------------------------------------------------
def get_peer_array(arr: Array, pe: int):
    """
    Returns an array view of a peer buffer associated with an NVSHMEM-allocated object.

    This is the Python array equivalent of `nvshmem_ptr` which returns a pointer into a peer's symmetric heap

    Args:
        array: Array - A symmetric array allocated by NVSHMEM
        pe: int - The remote PE to retrieve an ArrayView into

    Returns:
        - A Numba ArrayView which represents the peer object

    NOTE: This function is only supported with the Numba Runtime. To enable it,
          set the environment variable ``NUMBA_CUDA_ENABLE_NRT=1``
    """
    pass


@overload(get_peer_array)
def get_peer_array_ol(arr: Array, pe: int) -> Array:
    
    if arr == Array(dtype=int8, ndim=arr.ndim, layout=arr.layout):
        def impl(arr: Array, pe: int) -> Array:
            base_optr = _array_data_noneptr(arr)          # i8*
            pe32      = np.int32(pe)                      # match (none*, int32)
            peer_optr = ptr(base_optr, pe32)     # (uint8*, int32) -> none*
            peer_tptr = _noneptr_as_typedptr(peer_optr, arr)  # T*
            return carray(peer_tptr, arr.shape, dtype=int8)
        return impl
    
    elif arr == Array(dtype=int16, ndim=arr.ndim, layout=arr.layout):
        def impl(arr: Array, pe: int) -> Array:
            base_optr = _array_data_noneptr(arr)          # i8*
            pe32      = np.int32(pe)                      # match (none*, int32)
            peer_optr = ptr(base_optr, pe32)     # (uint8*, int32) -> none*
            peer_tptr = _noneptr_as_typedptr(peer_optr, arr)  # T*
            return carray(peer_tptr, arr.shape, dtype=int16)
        return impl
    
    elif arr == Array(dtype=int32, ndim=arr.ndim, layout=arr.layout):
        def impl(arr: Array, pe: int) -> Array:
            base_optr = _array_data_noneptr(arr)          # i8*
            pe32      = np.int32(pe)                      # match (none*, int32)
            peer_optr = ptr(base_optr, pe32)     # (uint8*, int32) -> none*
            peer_tptr = _noneptr_as_typedptr(peer_optr, arr)  # T*
            return carray(peer_tptr, arr.shape, dtype=int32)
        return impl
    
    elif arr == Array(dtype=int64, ndim=arr.ndim, layout=arr.layout):
        def impl(arr: Array, pe: int) -> Array:
            base_optr = _array_data_noneptr(arr)          # i8*
            pe32      = np.int32(pe)                      # match (none*, int32)
            peer_optr = ptr(base_optr, pe32)     # (uint8*, int32) -> none*
            peer_tptr = _noneptr_as_typedptr(peer_optr, arr)  # T*
            return carray(peer_tptr, arr.shape, dtype=int64)
        return impl
    
    elif arr == Array(dtype=uint8, ndim=arr.ndim, layout=arr.layout):
        def impl(arr: Array, pe: int) -> Array:
            base_optr = _array_data_noneptr(arr)          # i8*
            pe32      = np.int32(pe)                      # match (none*, int32)
            peer_optr = ptr(base_optr, pe32)     # (uint8*, int32) -> none*
            peer_tptr = _noneptr_as_typedptr(peer_optr, arr)  # T*
            return carray(peer_tptr, arr.shape, dtype=uint8)
        return impl
    
    elif arr == Array(dtype=uint16, ndim=arr.ndim, layout=arr.layout):
        def impl(arr: Array, pe: int) -> Array:
            base_optr = _array_data_noneptr(arr)          # i8*
            pe32      = np.int32(pe)                      # match (none*, int32)
            peer_optr = ptr(base_optr, pe32)     # (uint8*, int32) -> none*
            peer_tptr = _noneptr_as_typedptr(peer_optr, arr)  # T*
            return carray(peer_tptr, arr.shape, dtype=uint16)
        return impl
    
    elif arr == Array(dtype=uint32, ndim=arr.ndim, layout=arr.layout):
        def impl(arr: Array, pe: int) -> Array:
            base_optr = _array_data_noneptr(arr)          # i8*
            pe32      = np.int32(pe)                      # match (none*, int32)
            peer_optr = ptr(base_optr, pe32)     # (uint8*, int32) -> none*
            peer_tptr = _noneptr_as_typedptr(peer_optr, arr)  # T*
            return carray(peer_tptr, arr.shape, dtype=uint32)
        return impl
    
    elif arr == Array(dtype=uint64, ndim=arr.ndim, layout=arr.layout):
        def impl(arr: Array, pe: int) -> Array:
            base_optr = _array_data_noneptr(arr)          # i8*
            pe32      = np.int32(pe)                      # match (none*, int32)
            peer_optr = ptr(base_optr, pe32)     # (uint8*, int32) -> none*
            peer_tptr = _noneptr_as_typedptr(peer_optr, arr)  # T*
            return carray(peer_tptr, arr.shape, dtype=uint64)
        return impl
    
    elif arr == Array(dtype=float32, ndim=arr.ndim, layout=arr.layout):
        def impl(arr: Array, pe: int) -> Array:
            base_optr = _array_data_noneptr(arr)          # i8*
            pe32      = np.int32(pe)                      # match (none*, int32)
            peer_optr = ptr(base_optr, pe32)     # (uint8*, int32) -> none*
            peer_tptr = _noneptr_as_typedptr(peer_optr, arr)  # T*
            return carray(peer_tptr, arr.shape, dtype=float32)
        return impl
    
    elif arr == Array(dtype=float64, ndim=arr.ndim, layout=arr.layout):
        def impl(arr: Array, pe: int) -> Array:
            base_optr = _array_data_noneptr(arr)          # i8*
            pe32      = np.int32(pe)                      # match (none*, int32)
            peer_optr = ptr(base_optr, pe32)     # (uint8*, int32) -> none*
            peer_tptr = _noneptr_as_typedptr(peer_optr, arr)  # T*
            return carray(peer_tptr, arr.shape, dtype=float64)
        return impl
    
    elif arr == Array(dtype=float16, ndim=arr.ndim, layout=arr.layout):
        def impl(arr: Array, pe: int) -> Array:
            base_optr = _array_data_noneptr(arr)          # i8*
            pe32      = np.int32(pe)                      # match (none*, int32)
            peer_optr = ptr(base_optr, pe32)     # (uint8*, int32) -> none*
            peer_tptr = _noneptr_as_typedptr(peer_optr, arr)  # T*
            return carray(peer_tptr, arr.shape, dtype=float16)
        return impl
    
    return None


###
# The below are shim bindings for nvshmem_ptr and nvshmemx_mc_ptr. 
# We need them because Numbast and LLVM doesn't support passing void*s fully. 
# We exploit our knowledge of the fact that these APIs always have void* as a one-degree pointer (e.g. a void* not a void**)
# So it is safe to use them this way. However, this is not a general solution so we can't use this approach inside Numbast
# This means we have to ship the bindings as part of nvshmem.core as is below
# TODO: remove the following section if there's a general solution in Numbast for passing void*s
###


def ptr():
    pass

def _lower__Z11nvshmem_ptr_nbst(shim_stream, shim_obj):


    shim_raw_str = """
    extern "C" __device__ int
    _Z11nvshmem_ptr_nbst(void * &retval , void ** ptr, int* pe) {
        retval = nvshmem_ptr(*ptr, *pe);
        return 0;
    }
        """


    _Z11nvshmem_ptr_nbst = declare_device(
        '_Z11nvshmem_ptr_nbst',
        CPointer(void)(
            CPointer(CPointer(uint8)), CPointer(int32)
        )
    )
    

    def _Z11nvshmem_ptr_nbst_caller(arg_0, arg_1):
        return _Z11nvshmem_ptr_nbst(arg_0, arg_1)
    

    @lower(ptr, CPointer(uint8), int32)
    def impl(context, builder, sig, args):
        context.active_code_library.add_linking_file(shim_obj)
        shim_stream.write_with_key("_Z11nvshmem_ptr_nbst", shim_raw_str)

        val_types = [context.get_value_type(arg) for arg in sig.args]
        ptrs = [builder.alloca(val_type) for val_type in val_types]
        for ptr, ty, arg in zip(ptrs, sig.args, args):
            builder.store(arg, ptr, align=getattr(ty, "alignof_", None))

        return context.compile_internal(
            builder,
            _Z11nvshmem_ptr_nbst_caller,
            signature(CPointer(void), CPointer(CPointer(uint8)), CPointer(int32)),
            ptrs,
        )
    

_lower__Z11nvshmem_ptr_nbst(bindings._numbast.shim_stream, bindings._numbast.shim_obj)


@register
class _typing_ptr(ConcreteTemplate):
    key = globals()["ptr"]
    cases = [
        signature(CPointer(void), CPointer(uint8), int32),
    ]


register_global(ptr, types.Function(_typing_ptr))

def mc_ptr():
    pass



def _lower__Z11nvshmemx_mc_ptr_nbst(shim_stream, shim_obj):


    shim_raw_str = """
    extern "C" __device__ int
    _Z11nvshmemx_mc_ptr_nbst(void * &retval , int* team, void ** ptr) {
        retval = nvshmemx_mc_ptr(*team, *ptr);
        return 0;
    }
        """


    _Z11nvshmemx_mc_ptr_nbst = declare_device(
        '_Z11nvshmemx_mc_ptr_nbst',
        CPointer(void)(
            CPointer(int32), CPointer(CPointer(uint8))
        )
    )
    

    def _Z11nvshmemx_mc_ptr_nbst_caller(arg_0, arg_1):
        return _Z11nvshmemx_mc_ptr_nbst(arg_0, arg_1)
    

    @lower(mc_ptr, int32, CPointer(uint8))
    def impl(context, builder, sig, args):
        context.active_code_library.add_linking_file(shim_obj)
        shim_stream.write_with_key("_Z11nvshmemx_mc_ptr_nbst", shim_raw_str)

        val_types = [context.get_value_type(arg) for arg in sig.args]
        ptrs = [builder.alloca(val_type) for val_type in val_types]
        for ptr, ty, arg in zip(ptrs, sig.args, args):
            builder.store(arg, ptr, align=getattr(ty, "alignof_", None))

        return context.compile_internal(
            builder,
            _Z11nvshmemx_mc_ptr_nbst_caller,
            signature(CPointer(void), CPointer(int32), CPointer(CPointer(uint8))),
            ptrs,
        )
    

_lower__Z11nvshmemx_mc_ptr_nbst(bindings._numbast.shim_stream, bindings._numbast.shim_obj)


@register
class _typing_mc_ptr(ConcreteTemplate):
    key = globals()["mc_ptr"]
    cases = [
        signature(CPointer(void), int32, CPointer(uint8)),
    ]


register_global(mc_ptr, types.Function(_typing_mc_ptr))